//---------------------------------------------------------------
//---------------------------------------------------------------
//------------  Simple Genetic Algorithm             ------------
//------------  - function optimization
//------------  - binary representation              ------------  
//------------  - generational replacement strategy  ------------  
//---------------------------------------------------------------
//---------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <conio.h>
#include <time.h>
#include <string.h>
#include <float.h>
//----------------

//----------------------------------------------------------------------
//----------------------------------------------------------------------

//------- individual structure definition
typedef struct {
  int *genes;		//--- binary vector
  double fitness;
} tind;

//------- statistic parameters observed
typedef struct {
  double fbest;   // best fitness in the population (best-so-far)
  double favg;    // average fitness of the chromosomes
  double fworst;  // fitness of the worst chromosome
  double fsum;    // sum of fitness values over the whole population
  int indbest;    // index of the best individual
  int indworst;   // index of the worst individual
} tstat;
tstat statdata;

//----------------------------------------------------------------------
//----------------------------------------------------------------------
int MaxGenes = 24;				// fixed length of chromosomes
int PopSize = 100;				// population size
int Evaluations = 10000;		// stopping criterion
int crossType = 0;				// crossover type 0...uniform, 1...2-point
int select_1= 3, select_2= 3;	// parameters of tournament selection
int Pc = 75;					// probability of crossover [%]
int Pm = 5;						// probability of mutation [%]
int rand_seed = 1;				// seeds for pseudo-random generator
double minX = 0.0;
double maxX = 50.0;
double rozsahX = maxX - minX;

//---------------------------
char *name_output = "sga_func_opt.vys\0";  // output file
FILE *report_file;

//---------------------------
tind *oldpop, *newpop, *backpop, *offspring; // old and new population


//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------- decode genotype stored in 'genes'
//------- return an integer value coded by binary genes
double decode(int *genes,int from,int to){
	int i;
	int mocnina=1;
	double pom=0.0;

	for(i=from;i<to;i++){
		pom+= (mocnina*genes[i]);
		mocnina*=2;
	};

	return pom;
}

//----------------------------------------------------------------------
//------- Fitness Function Definition
//------- f(x) = sin(pi * x) + 10*sin(pi * x / rozsahX)
double calculate_fitness(double x){
	double pi = 3.1415926535;

	//------- calculate f(x)
	return (sin(pi * x) + 10.0 * sin(pi * x/rozsahX));
}

//----------------------------------------------------------------------
//------- Fitness Function calculation
void get_fitness(tind &individual){
	double pom = pow(2.0, MaxGenes) - 1.0;
	double x;

	//------- decode parameter x
	x = decode(individual.genes, 0, MaxGenes);
	x = minX + (x / pom) * rozsahX;

	//------- calculate f(x)
	individual.fitness = calculate_fitness(x);
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------- overall initialisation
//------- allocates memory for individuals
void init_all(int init_seed){
	int i,j;

	srand(init_seed);

	//------- allocate memory for chromosomes
	for(j=0;j<PopSize;j++){
		oldpop[j].genes = (int *)malloc(MaxGenes*sizeof(int));
		newpop[j].genes = (int *)malloc(MaxGenes*sizeof(int));
	};
	offspring[0].genes = (int *)malloc(MaxGenes*sizeof(int));
	offspring[1].genes = (int *)malloc(MaxGenes*sizeof(int));

	//------- initialize chromosomes
	//------- randomly generates chromosomes with probability of generating 1 of P(1)=0.5 
	for(i=0; i<PopSize; i++){ 
		//------- 
		for(j=0; j<MaxGenes; j++){
			if(rand()%100 < 50){              
				oldpop[i].genes[j] = 1;
			}
			else{
				oldpop[i].genes[j] = 0;
			}
		};
		//------- send the GA to poor region
		if(rand() % 2)
			for(j=MaxGenes-1; j>=MaxGenes/2; j--)
				oldpop[i].genes[j] = 0;
		else
			for(j=MaxGenes-1; j>=MaxGenes/2; j--)
				oldpop[i].genes[j] = 1;
	};

	//------- evaluate chromosomes
	for(i=0;i<PopSize;i++){
		get_fitness(oldpop[i]);
	};
};

  
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------- population statistics
void statistics(tstat *statdata,tind *pop){
	int i;
	double pom=0.0;

	//------- init statdata
	statdata->fbest = -10000.0;
	statdata->fworst = 10000;
	statdata->favg = 0.0;
	statdata->fsum = 0.0;
	statdata->indbest = 0;
	statdata->indworst = 0;
	for(i=0;i<PopSize;i++){
		if(statdata->fbest<=pop[i].fitness){statdata->fbest=pop[i].fitness;statdata->indbest=i;};
		if(statdata->fworst>=pop[i].fitness){statdata->fworst=pop[i].fitness;statdata->indworst=i;};
		statdata->fsum+=pop[i].fitness;
	};
	statdata->favg = statdata->fsum / PopSize;
};

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------- reports results
void report_results(tstat *statdata,tind *pop, int evals){
  
  printf("(%d)	fbest: %.10f 	favg: %.10f\n", evals, statdata->fbest, statdata->favg);
};

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------- saves population stats
void save_results(tstat *statdata,tind *pop, int evals){
  int j=1;

  fprintf(report_file, "\n%d	%.4f	%.4f	%.4f", evals,statdata->fbest, statdata->favg, statdata->fworst);
};

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------- saves the best chromosome
void save_best(int evals, tstat *statdata,tind *pop){
	double pom = pow(2.0, MaxGenes) - 1.0; 
	double x;

	//------- decode parameters x
	x = decode(pop[statdata->indbest].genes, 0, MaxGenes);  // get the decimal value of the first 17 bits
	x = minX + (x / pom) * rozsahX;

	//------- print the values of x
	fprintf(report_file,"\n%.10f	%.10f", pop[statdata->indbest].fitness, x);
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------- copies the individual from pop_from to pop_to
void copy_ind(int ind_from,tind *pop_from,int ind_to,tind *pop_to){
  int i;

  for(i=0;i<MaxGenes;i++){
	  pop_to[ind_to].genes[i] = pop_from[ind_from].genes[i];
  };
  pop_to[ind_to].fitness = pop_from[ind_from].fitness;
}

//----------------------------------------------------------------------
//------- crossover - 2-point classical with mutation
void cross_2point(int *parent1, int *parent2, int *child1,int *child2){

	int i;
	int lp,rp,pom;

	//------- choose two crossing points - left and right
	lp=rand()%MaxGenes;
	do rp=(rand()%MaxGenes); while(lp==rp);
	if(lp>rp){
		pom=lp;
		lp=rp;
		rp=pom;
	}

	//------- head section
	for(i=0;i<lp;i++){
		child1[i] = parent1[i];
		child2[i] = parent2[i];
	}

	//------- middle section
	for(i=lp;i<rp;i++){
		child2[i] = parent1[i];
		child1[i] = parent2[i];
	}

	//------- end section
	for(i=rp;i<MaxGenes;i++){
		child1[i] = parent1[i];
		child2[i] = parent2[i];
	}
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------- crossover - uniform
void cross_uniform(int *parent1, int *parent2, int *child1,int *child2){

	int i;

	//------- each gene is inherited either from the 1st or the 2nd parent
	for(i=0; i<MaxGenes; i++){
		//------- child1
		if(rand() % 2)
			child1[i] = parent1[i];
		else
			child1[i] = parent2[i];
		//------- child2
		if(rand() % 2)
			child2[i] = parent1[i];
		else
			child2[i] = parent2[i];
	}
}

//----------------------------------------------------------------------
//------- crossover
void cross(int crossType, int *parent1, int *parent2, int *child1,int *child2){

	if(crossType == 0)
		cross_uniform(parent1, parent2, child1, child2);
	else
		cross_2point(parent1, parent2, child1, child2);
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------- Mutation - simple bit-flip mutation
void mutate(int *chromosome){
	int i;

	//------- Applies mutation to every gene with probability Pm
	for(i=0;i<MaxGenes;i++)
		if(rand()%100 < Pm)
			chromosome[i] = (chromosome[i] + 1) % 2;	//--- changes 1->0 and 0->1
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------ Tournament selection                   
//------   - tournamentSize individuals take part in tournament
int select_tournament(tind *pop, int tournamentSize){
	int cand, tempbest;

	//------- 1st competitor
	tempbest = rand()%PopSize; //--- best-so-far candidate
	tournamentSize--;

	//------- try remaining competitors
	while(tournamentSize > 0){ 
		cand = rand() % PopSize; 
		if(pop[cand].fitness > pop[tempbest].fitness)
			tempbest = cand;
		//-------
		tournamentSize--;
	}
	return tempbest;
};

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------ Chooses parents using select() strategy                   
void choose_parents(tind *pop, int &par1, int &par2){

	//------- 1st parent
	par1 = select_tournament(pop,select_1);

	//------- chooses 2nd parent to be different than the 1st one
	do{
		par2 = select_tournament(pop, select_2);
	}while(par1 == par2);
}

//------- Reads parameters from command line
void read_parameters(int argc, char *argv[]){

	if(argc == 8){
		PopSize = atoi(argv[1]);			//--- population size
		Evaluations = atoi(argv[2]);		//--- number of fitness evaluations
		crossType = atoi(argv[3]);			//--- crossover type
		Pc = atoi(argv[4]);					//--- treti argument je Pc
		Pm = atoi(argv[5]);					//--- Pm
		select_1 = select_2 = atoi(argv[6]);//--- tournament parameter
		rand_seed = atoi(argv[7]);			//--- rand_seed
	}
	else{
		printf("\nWrong parameters");
		printf("\nExpected parameters: PopSize Evals Crossover Pcross Pmut Tournament Seed");
		printf("\nPopSize - population size [positive integer]");
		printf("\nEvals - maximal number of fitness evaluations [positive integer]");
		printf("\nCrossover - crossover type: 0...uniform, 1...2-point");
		printf("\nPcross - probability of crossover: 0-100 %");
		printf("\nPmut - probability of mutation: 0-100 %");
		printf("\nTournament - parameter of tournament selection [positive integer]");
		printf("\nSeed - random number generator initialization value [positive integer]");
	}
}

//--------------------------------------------------------
//-------------------- MAIN EVOLUTION --------------------
//--------------------------------------------------------
int main(int argc,char *argv[]){
	int TotEvals;		    // a number of evaluations made so far
	int cgen;			    // current generation	
	int par1, par2;         // indexes of parents
	int chroms_generated;   // a number of newly generated chromosomes
	int report_step = 2000; // a number of evaluations between two reports
	int report_count = 0;
	int i;

	//-------
	read_parameters(argc, argv);
	report_file = fopen(name_output,"w"); 

	//------------------------------------------------
	//------- Initialization of the genetic algorithm
	//------------------------------------------------
	cgen=0;
	TotEvals = 0;
	report_count = 0;
	
	//------- allocate memory 
	oldpop = (tind *)malloc(PopSize*sizeof(tind));
	newpop = (tind *)malloc(PopSize*sizeof(tind));
	offspring = (tind *)malloc(2*sizeof(tind));

	//-------
	init_all(rand_seed);
	statistics(&statdata, oldpop);

	//-------
	report_results(&statdata,newpop,TotEvals);
	save_best(TotEvals, &statdata,oldpop);

	//-------------------------------------------
	//------- Main loop of the genetic algorithm
	//-------------------------------------------
	do{//------- generate new population
		chroms_generated = 0;
		cgen++;

		//------- elitism
		copy_ind(statdata.indbest, oldpop, chroms_generated++, newpop); 

		//------- create new population of chromosomes (generational replacement strategy)
		do{
			choose_parents(oldpop, par1, par2);	// choose parents
			//------- create two new chromosomes
			if(rand()%100 < Pc){
				cross(crossType, oldpop[par1].genes, oldpop[par2].genes, offspring[0].genes, offspring[1].genes);
				mutate(offspring[0].genes);
				mutate(offspring[1].genes);
			}
			else{
				copy_ind(par1, oldpop, 0, offspring);
				copy_ind(par2, oldpop, 1, offspring);
				mutate(offspring[0].genes);
				mutate(offspring[1].genes);
			}
			//------- evaluate offspring
			get_fitness(offspring[0]); TotEvals++;
			get_fitness(offspring[1]); TotEvals++;
			//------- replacement
			copy_ind(0, offspring, chroms_generated++, newpop);
			if(chroms_generated < PopSize)
				copy_ind(1, offspring, chroms_generated++, newpop);
		}while(chroms_generated < PopSize);

		//------- switch populations
		backpop = oldpop;
		oldpop = newpop;
		newpop = backpop;

		//-------
		statistics(&statdata, oldpop);
		save_best(TotEvals, &statdata, oldpop);
		report_results(&statdata, oldpop,TotEvals); 
	}while(TotEvals < Evaluations);

	//-------
	report_results(&statdata,newpop,TotEvals); 
	save_best(TotEvals, &statdata, oldpop);

	//------ free memory
	for(i=0;i<PopSize;i++){
		free(oldpop[i].genes);
		free(newpop[i].genes);
	};
	free(offspring[0].genes); free(offspring[1].genes);
	free(oldpop); 
	free(newpop); 
	free(offspring);
	fclose(report_file);

	//-------
	printf("\n\nFinished ...");
	getch();
	return 0;
}
